import os
import abc
from enum import Enum
from typing import final
from dataclasses import dataclass

import torch
from torch.utils import data

from .utils import Utils
from .platform import PlatformConfig
from .artefacts import ArtefactsConfig


@final
class JobMode(Enum):
  r"""Modes of running the task."""

  TRAIN = 0
  r"""The task is to train."""

  RESUME = 1
  r"""The task is to resume training from a checkpoint."""

  EVALUATE = 2
  r"""The task is to evaluate a model checkpoint (`i.e.`, a trained model)."""


@final
@dataclass
class JobConfig(object):
  job_type: JobMode = JobMode.TRAIN
  r"""Type of job. Default: ``JobMode.TRAIN``."""

  start_at: int = 0
  r"""
  Epoch number from which to start training. Default: ``0``.

  Setting this to a negative value skips model/state restoration in case of
  testing or restoration. This is useful if there it no model to perform the
  task on, _i.e._, this is useful for running jobs that are to be done on the
  GPU that are not necessarily training or evaluating a model.
  """

  epochs: int = 25
  r"""Epoch number at which to stop training. Default: ``25``."""

  checkpoint_name_prefix: str = 'ckpt'
  r"""Prefix for saving the checkpoint. Default: ``ckpt``."""

  checkpoint_path: str = './checkpoint'
  r"""Location at which to store the checkpoint. Default: ``./checkpoint``."""

  console_logs_path: str = './logs/console_logs'
  r"""Location at which to store console logs. (`e.g.`, logs generated by
    SLURM). Default: ``./logs/console_logs``."""

  training_logs_path: str = './logs/training_logs'
  r"""Location at which to store training logs (`e.g.`, logs from Tensorboard).
  Default: ``./logs/training_logs``."""

  save_every: int = 5
  r"""Save a checkpoint every few epochs. If 0, ignored. Default: ``5``."""

  def print(self):
    r"""
    This method prints this object in a readable format.
    """

    Utils.print('Job details:')
    Utils.print(f' • Job type:                            {self.job_type}')
    Utils.print(f' • Starting epoch:                      {self.start_at}')
    Utils.print(f' • Epochs:                              {self.epochs}')
    Utils.print(' • State/checkpoint save location:      ' +
                f'{self.checkpoint_path}')
    Utils.print(' • State/checkpoint name prefix:        ' +
                f'{self.checkpoint_name_prefix}')
    Utils.print(' • Console logs path:                   ' +
                f'{self.console_logs_path}')
    Utils.print(' • Training logs path:                  ' +
                f'{self.training_logs_path}')
    Utils.print(' • Save state/models every...           ' +
                f'{self.save_every} epochs')


@dataclass
class Job(object):
  r"""
  This is a template class with abstract methods to be defined by the user. This
  class provides methods to define training and evaluation procedures. Once the
  wrapper has moved the model and the dataset to the appropriate device, it
  calls :py:meth:`.train()` or :py:meth:`.evaluate()` as configured.
  """

  p_config: PlatformConfig = None
  r"""
  Platform-related configuration. This property may be used in the training
  and evaluation methods to access platform-related information such as if the
  platform is on CUDA, how big the world is, whether synchronisation across
  devices is needed, `etc`.

  .. admonition:: Definition not required
   :class: note

   This property need not be specified by the user and will be automatically
   updated by the wrapper right before training or evaluation. This can be
   directly accessed in the :py:meth:`~Job.train` and
   :py:meth:`~Job.evaluate` methods.
  """

  j_config: JobConfig = None
  r"""
  Training-related configurations. This property may be used in the training
  and evaluation methods to access training-specific aspects such as the number
  of training epochs, epoch interval to store the training state, `etc`.
  """

  a_config: ArtefactsConfig = None
  r"""
  Model-related configuration. This property may be used in the training and
  evaluation methods to access models, datasets, optimisation stragegy (the
  optimiser) `etc.`

  .. admonition:: Definition not required
   :class: note

   This property need not be specified by the user and will be automatically
   updated by the wrapper right before training or evaluation. This can be
   directly accessed in the :py:meth:`~Job.train` and
   :py:meth:`~Job.evaluate` methods.
  """

  @abc.abstractmethod
  def train(self, global_rank: int, local_rank: int):
    r"""
    .. admonition:: Definition required
      :class: important

      This method needs to be explicitly defined by the user.

    This method provides definition for the training procedure.

    :param int global_rank: Global rank of the current device.
    :param int local_rank: Local rank of the current device.

    :raises NotImplementedError: Training has not been implemented.
    """

    raise NotImplementedError

  @abc.abstractmethod
  def evaluate(self, global_rank: int, local_rank: int,
               dataset: data.DataLoader):
    r"""
    .. admonition:: Definition required
      :class: important

      This method needs to be explicitly defined by the user.

    This method provides definition for the evaluation procedure.

    :param int global_rank: Global rank of the current device.
    :param int local_rank: Local rank of the current device.
    :param data.DataLoader dataset: The dataset to use for evaluation.

    :raises NotImplementedError: Evaluation has not been implemented.
    """

    raise NotImplementedError

  def save_state(self, epoch: int):
    r"""
    This method saves the state of the training, such as the model parameters,
    optimiser gradients, the current epoch, `etc`., and can be called at every
    few epochs (as specified in :py:attr:`.JobConfig.save_every`) or as
    needed. Override this method to save more information. The state so saved is
    used by the :py:meth:`.restore_state` method that is called to resume from
    a checkpoint or evaluate a model.

    :param int epoch: The epoch number at which to save the training state.
    """

    checkpoint = {
      'stopped_at': epoch,
      'model': self.a_config.model.state_dict(),
      'optimiser': self.a_config.optimiser.state_dict()
    }
    torch.save(checkpoint, os.path.join(self.j_config.checkpoint_path,
                          f'{self.j_config.checkpoint_name_prefix}_{epoch}.pt'))

  def restore_state(self, resume_at: int):
    r"""
    Restore training from a saved state.

    :param int resume_at: The epoch checkpoint whence to resume training.
    """

    filename = f'{self.j_config.checkpoint_name_prefix}_{resume_at}.pt'
    file_path = os.path.join(self.j_config.checkpoint_path, filename)
    print(file_path)
    assert os.path.isfile(file_path)

    Utils.print(f'Loading model at {file_path}.')
    checkpoint = torch.load(file_path)
    self.j_config.start_at = checkpoint['stopped_at']
    self.a_config.model.load_state_dict(checkpoint['model'])
    self.a_config.optimiser.load_state_dict(checkpoint['optimiser'])

  def __call__(self, global_rank: int, local_rank: int):
    r"""
    When once the distributed data parallel setups are completed by the wrapper,
    this method is called. This method locally updates the dataset and model
    allotted for the current GPU in case of GPU- and SLURM-based platforms.

    :param int global_rank: The global rank of the device.
    :param int local_rank: Local rank of the current device.
    """

    Utils.print(
      f'[Device {global_rank}] Copying model parameters to the optimiser.')
    if self.a_config.optimiser_config is not None:
      self.a_config.optimiser = self.a_config.optimiser_config(
        self.a_config.model)

    # if this task is resumption from or evaluation of a saved model, load it
    if self.j_config.job_type in [JobMode.RESUME, JobMode.EVALUATE]:
      Utils.print(f'[Device {global_rank}] Model load setup underway.')
      if self.j_config.start_at >= 0:
        self.restore_state(self.j_config.start_at)

    # whether to training (or resumption) or evaluate
    if self.j_config.job_type in [JobMode.TRAIN, JobMode.RESUME]:
      self.train(global_rank, local_rank)
    else:
      self.evaluate(global_rank, local_rank, self.a_config.test_set)



import abc
from typing import final, Optional, Callable
from dataclasses import dataclass

from torch import nn, optim
from torch.utils import data

from .utils import Utils


class OptimiserLoader(object):
  r"""
  This class allows usage of optimisers with desired configurations. Before it
  begins the training/evaluation process, it uses the optimiser specified in
  this class.

  Of course, this class could be extended to customise as desired: to use
  different optimiser-specific parameters such as the learning rate, weight
  decay, momentum, `etc.`, as well as defining a set of optimisers to choose
  from depending on the desired optimiser at run-time.
  """

  @abc.abstractmethod
  def __call__(self, model: Optional[nn.Module]) -> optim.Optimizer:
    r"""
    .. admonition:: Definition required
      :class: important

      This method needs to be explicitly defined by the user.

    This method receives the model whose parameters are to be loaded into the
    optimiser. Once configured as desired, it returns the optimiser.

    :param Optional[nn.Module] model: The model whose parameters are to be
      loaded into the optimiser.
    :returns optim.Optimizer: The desired optimiser with configurations set and
        model parameters loaded
    :raises NotImplementedError: Method not implemented.
    """

    raise NotImplementedError


@final
@dataclass
class ArtefactsConfig(object):
  r"""Configurations relating to the dataset and the model."""

  train_set: data.Dataset = None
  r"""The dataset to use for training. Default: ``None``."""

  test_set: data.Dataset = None
  r"""The dataset to use for evaluation. Default: ``None``."""

  validation_set: data.Dataset = None
  r"""The dataset to use for validation. Default: ``None``."""

  collate_fn: Optional[Callable] = None
  r"""Any callable function to be passed down to the dataloader. Default:
   ``None``."""

  validation_percentage: float = 0
  r"""The percentage of training dataset to be used for validation. If this
  property has a value of ``0``, it is assumed that no validation is required.
  This property is ignored for evaluation. Range: ``0`` to ``50``, inclusive.
  Default: ``0``. Range enforced by the :py:attr:`.needs_validation`
  property."""

  batch_size: int = 64
  r"""Batch size for training and testing. Default: ``64``."""

  model: nn.Module = None
  r"""An instance of the model to train. Default: ``None``."""

  model_has_batch_norm: bool = False
  r"""Specifies if the model to be trained has batch normalisation in it.
  Default: ``False``."""

  optimiser_config: OptimiserLoader = None
  r"""Optimiser setup to be passed by the user. Default: ``None``."""

  optimiser: optim.Optimizer = None
  r"""
  .. admonition:: Definition not required
   :class: note

   This property need not be specified by the user and will be automatically
   updated by the wrapper right before training. This can be directly in the
   :py:meth:`.Job.train` method.

  The wrapper loads model parameters into the optimiser with the specified
  configs in :py:attr:`optimiser_config` and updates this parameter. This could
  be accessed in :py:meth:`.Job.train` or :py:meth:`.Job.evaluate`.
  Default: ``None``.
  """

  @property
  def needs_validation(self):
    r"""
    This property tells if the current configuration requires validation or not.

    :returns bool: Whether validation is required or not.
    """

    return 0 < self.validation_percentage <= 50

  def print(self):
    r"""
    This method prints this object in a readable format.
    """

    Utils.print('Artefacts details:')
    Utils.print(' • Train dataset:                       ' +
                f'{len(self.train_set) if self.train_set is not None else 0}')
    Utils.print(' • Test dataset:                        ' +
                f'{len(self.test_set) if self.test_set is not None else 0}')
    if self.needs_validation:
      Utils.print(' • Validation percentage:               ' +
                  f'{self.validation_percentage}')
    Utils.print(f' • Batch size:                          {self.batch_size}')



from typing import Optional, Callable

from torch.utils import data
from torch.utils.data import DistributedSampler, DataLoader, random_split

from ..utils import Utils
from ..artefacts import ArtefactsConfig
from ..platform import Platform, PlatformConfig


def sampler(dataset: data.Dataset, world_size: int, global_rank: int,
            batch_size: int, collate_fn: Optional[Callable] = None,
            is_cpu: bool = False):
  r"""
  This function selects a portion of the original dataset shared by other
  devices. If the device being trained on is a CPU, no sharing is necessary.

  :param data.Dataset dataset: The dataset from which to sample for the current
    device.
  :param int world_size: World size.
  :param int global_rank: Global rank of the current GPU.
  :param int batch_size: Batch size.
  :param Optional[Callable] collate_fn: The collate function to use in the
    dataloader.
  :param bool is_cpu: Specifies if the device is a CPU or Apple M1. Default:
    `False`.

  :returns data.Dataset: A dataloader with portion of the dataset selected for
      the current process.
  """

  smplr = None if is_cpu else DistributedSampler(dataset, world_size,
                                                 rank=global_rank)
  return DataLoader(dataset, batch_size, sampler=smplr, pin_memory=True,
                    collate_fn=collate_fn)


def dataset_setup(global_rank: int, p_config: PlatformConfig,
                  a_config: ArtefactsConfig):
  r"""
  This function selects a portion of the training dataset for validation if
  specified. In case of training/testing on multiple devices, it then allocates
  a portion each of the training, test, and validation datasets (if available)
  to the current device and returns a dataloader for each.

  :param int global_rank: Global rank of the current GPU.
  :param PlatformConfig p_config: Platform configurations.
  :param ArtefactsConfig a_config: Job configurations.

  :returns tuple: A triplet of dataloaders for the training, validation, and
      test datasets respectively.
  """

  train_loader = None
  val_loader = None
  test_loader = None

  val_set = a_config.validation_set

  is_cpu = p_config.platform in [Platform.CPU, Platform.MPS]
  batch_size, collate_fn = a_config.batch_size, a_config.collate_fn

  args = (p_config.world_size, global_rank, batch_size, collate_fn, is_cpu)

  # if the training dataset is provided
  if (train_set := a_config.train_set) is not None:

    # if requested to set aside a portion of the training set for validation
    if a_config.needs_validation:
      dataset_size = len(train_set)
      v_size = (dataset_size * a_config.validation_percentage) // 100
      t_size = dataset_size - v_size
      Utils.print(f'\tTrain size = {t_size}; validation size = {v_size}.')
      [train_set, val_set] = random_split(train_set, [t_size, v_size])
    if val_set is not None:
      val_loader = sampler(val_set, *args)

    train_loader = sampler(train_set, *args)

  # if the test dataset is provided
  if (test_set := a_config.test_set) is not None:
    Utils.print(f'\tTest size  {len(test_set)}.')
    test_loader = sampler(test_set, *args)

  return train_loader, val_loader, test_loader
